@@ -3,13 +3,14 @@
     Author: Tyler Barrus (barrust@gmail.com)
     URL: https://github.com/barrust/bloom
 """
-
 import mmap
 import os
-import typing
+from array import array
+from collections.abc import ByteString
 from pathlib import Path
 from shutil import copyfile
-from struct import calcsize, pack, unpack
+from struct import Struct
+from typing import Union
 
 from ..exceptions import InitializationError, NotSupportedError
 from ..hashes import HashFuncT, HashResultsT
@@ -18,7 +19,7 @@ from .basebloom import BaseBloom
 
 MISMATCH_MSG = "The parameter second must be of type BloomFilter or a BloomFilterOnDisk"
 
-SimpleBloomT = typing.Union["BloomFilter", "BloomFilterOnDisk"]
+SimpleBloomT = Union["BloomFilter", "BloomFilterOnDisk"]
 
 
 def _verify_not_type_mismatch(second: SimpleBloomT) -> bool:
@@ -75,34 +76,33 @@ def _tmp_intersection(first: SimpleBloomT, second: SimpleBloomT) -> "BloomFilter
 
 
 class BloomFilter(BaseBloom):
-    """ Simple Bloom Filter implementation for use in python;
-        It can read and write the same format as the c version
-        (https://github.com/barrust/bloom)
+    """Simple Bloom Filter implementation for use in python;
+    It can read and write the same format as the c version
+    (https://github.com/barrust/bloom)
 
-        Args:
-            est_elements (int): The number of estimated elements to be added
-            false_positive_rate (float): The desired false positive rate
-            filepath (str): Path to file to load
-            hex_string (str): Hex based representation to be loaded
-            hash_function (function): Hashing strategy function to use \
-            `hf(key, number)`
-        Returns:
-            BloomFilter: A Bloom Filter object
-        Note:
-            Initialization order of operations:
-                1) From file
-                2) From Hex String
-                3) From params """
+    Args:
+        est_elements (int): The number of estimated elements to be added
+        false_positive_rate (float): The desired false positive rate
+        filepath (str): Path to file to load
+        hex_string (str): Hex based representation to be loaded
+        hash_function (function): Hashing strategy function to use `hf(key, number)`
+    Returns:
+        BloomFilter: A Bloom Filter object
+    Note:
+        Initialization order of operations:
+            1) From file
+            2) From Hex String
+            3) From params"""
 
     __slots__ = BaseBloom.__slots__
 
     def __init__(
         self,
-        est_elements: typing.Optional[int] = None,
-        false_positive_rate: typing.Optional[float] = None,
-        filepath: typing.Optional[str] = None,
-        hex_string: typing.Optional[str] = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        est_elements: Union[int, None] = None,
+        false_positive_rate: Union[float, None] = None,
+        filepath: Union[str, Path, None] = None,
+        hex_string: Union[str, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
         """setup the basic values needed"""
         super(BloomFilter, self).__init__(
@@ -114,6 +114,25 @@ class BloomFilter(BaseBloom):
             hash_function,
         )
 
+    __HEADER_STRUCT_FORMAT = "QQf"
+    __HEADER_STRUCT = Struct(__HEADER_STRUCT_FORMAT)
+    # __HEADER_STRUCT_BE = Struct(">" + __HEADER_STRUCT_FORMAT)
+
+    @classmethod
+    def frombytes(cls, b: ByteString, hash_function: Union[HashFuncT, None] = None) -> "BloomFilter":
+        """
+        Args:
+            b (ByteString): The bytes to load as a Bloom Filter
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            BloomFilter: A Bloom Filter object
+        """
+        offset = cls.__HEADER_STRUCT.size
+        est_elms, _, fpr, _, _, _ = cls._parse_footer(cls.__HEADER_STRUCT, bytes(b[-offset:]))
+        blm = BloomFilter(est_elements=est_elms, false_positive_rate=fpr, hash_function=hash_function)
+        blm._load(b, hash_function=blm.hash_function)
+        return blm
+
     def __str__(self) -> str:
         """output statistics of the bloom filter"""
         on_disk = "no" if self.is_on_disk is False else "yes"
@@ -145,7 +164,7 @@ class BloomFilter(BaseBloom):
             on_disk,
         )
 
-    def intersection(self, second: SimpleBloomT) -> typing.Optional["BaseBloom"]:
+    def intersection(self, second: SimpleBloomT) -> Union["BaseBloom", None]:
         """ Return a new Bloom Filter that contains the intersection of the
             two
 
@@ -170,7 +189,7 @@ class BloomFilter(BaseBloom):
 
         return _tmp_intersection(self, second)
 
-    def union(self, second: SimpleBloomT) -> typing.Optional["BaseBloom"]:
+    def union(self, second: SimpleBloomT) -> Union["BaseBloom", None]:
         """ Return a new Bloom Filter that contains the union of the two
 
             Args:
@@ -194,7 +213,7 @@ class BloomFilter(BaseBloom):
 
         return _tmp_union(self, second)
 
-    def jaccard_index(self, second: SimpleBloomT) -> typing.Optional[float]:
+    def jaccard_index(self, second: SimpleBloomT) -> Union[float, None]:
         """ Calculate the jaccard similarity score between two Bloom Filters
 
             Args:
@@ -248,15 +267,15 @@ class BloomFilterOnDisk(BaseBloom):
                 2) From Hex String
                 3) Only filepath provided """
 
-    __slots__ = ["__file_pointer", "__filename", "__export_offset"]
+    __slots__ = ["__file_pointer", "__filename"]
 
     def __init__(
         self,
-        filepath: typing.Union[str, Path],
-        est_elements: typing.Optional[int] = None,
-        false_positive_rate: typing.Optional[float] = None,
-        hex_string: typing.Optional[str] = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        filepath: Union[str, Path],
+        est_elements: Union[int, None] = None,
+        false_positive_rate: Union[float, None] = None,
+        hex_string: Union[str, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
         # since we cannot load from a file only (to memory), we can't pass
         # the file to the constructor; therefore, we will have to catch
@@ -273,34 +292,45 @@ class BloomFilterOnDisk(BaseBloom):
 
         self.__file_pointer = None
         self.__filename = Path(filepath)
-        self.__export_offset = calcsize("Qf")
         self._on_disk = True
 
         if est_elements is not None and false_positive_rate is not None:
             # no need to check the file since this will over write it
             fpr = false_positive_rate
-            vals = super(BloomFilterOnDisk, self)._set_optimized_params(est_elements, fpr, hash_function)
+            h_func, fpr, _, _ = self._set_optimized_params(est_elements, fpr, hash_function)
             super(BloomFilterOnDisk, self).__init__(
                 "reg-ondisk",
                 est_elements=est_elements,
-                false_positive_rate=vals[1],
-                hash_function=vals[0],
+                false_positive_rate=fpr,
+                hash_function=h_func,
             )
             # do the on disk things
             with open(filepath, "wb") as filepointer:
-                for _ in range(self.bloom_length):
-                    filepointer.write(pack("B", int(0)))
-                filepointer.write(pack("QQf", est_elements, 0, false_positive_rate))
+                (array("B", [0]) * self.bloom_length).tofile(filepointer)
+                filepointer.write(self.CNT_FOOTER_STUCT.pack(est_elements, 0, false_positive_rate))
                 filepointer.flush()
-            self.__load(filepath, hash_function)
+            self._load(filepath, hash_function)
         elif is_hex_string(hex_string):
             self._load_hex(hex_string, hash_function)
         elif is_valid_file(filepath):
-            self.__load(filepath, hash_function)
+            self._load(filepath, hash_function)
         else:
             msg = "Insufecient parameters to set up the On Disk Bloom Filter"
             raise InitializationError(msg)
 
+    CNT_ELM_STRUCT = Struct("B")
+    CNT_EXP_ELM_STRUCT = Struct("Q")
+    CNT_FOOTER_STUCT = Struct("QQf")
+    CNT_EXPORT_OFFSET = Struct("Qf")
+
+    @classmethod
+    def frombytes(cls, b: ByteString, hash_function: Union[HashFuncT, None] = None) -> "BloomFilter":
+        """
+        Raises: NotSupportedError
+        """
+        msg = "Loading from bytes is currently not supported by the on disk Bloom Filter"
+        raise NotSupportedError(msg)
+
     def __del__(self) -> None:
         """handle if user doesn't close the on disk Bloom Filter"""
         self.close()
@@ -312,31 +342,31 @@ class BloomFilterOnDisk(BaseBloom):
         """Clean up the BloomFilterOnDisk object"""
         if self.__file_pointer is not None:
             self.__update()
-            self._bloom.close()  # close the mmap
+            self._bloom.close()  # type: ignore
             self.__file_pointer.close()
             self.__file_pointer = None
 
-    def __load(self, filepath: typing.Union[str, Path], hash_function: typing.Optional[HashFuncT] = None):
+    def _load(self, filepath: Union[str, Path], hash_function: Union[HashFuncT, None] = None):  # type: ignore
         """load the Bloom Filter on disk"""
         # read the file, set the optimal params
         # mmap everything
         with open(filepath, "r+b") as filepointer:
-            offset = calcsize("QQf")
+            offset = self.CNT_FOOTER_STUCT.size
             filepointer.seek(offset * -1, os.SEEK_END)
-            mybytes = unpack("QQf", filepointer.read(offset))
-            vals = super(BloomFilterOnDisk, self)._set_optimized_params(mybytes[0], mybytes[2], hash_function)
+            est_els, _, fpr = self.CNT_FOOTER_STUCT.unpack_from(filepointer.read(offset))
+            hash_func, fpr, _, _ = self._set_optimized_params(est_els, fpr, hash_function)
         super(BloomFilterOnDisk, self).__init__(
             "reg-ondisk",
-            est_elements=mybytes[0],
-            false_positive_rate=vals[1],
-            hash_function=vals[0],
+            est_elements=est_els,
+            false_positive_rate=fpr,
+            hash_function=hash_func,
         )
         self.__file_pointer = open(filepath, "r+b")  # type: ignore
         self._bloom = mmap.mmap(self.__file_pointer.fileno(), 0)  # type: ignore
         self._on_disk = True
         self.__filename = Path(filepath)
 
-    def export(self, filename: typing.Union[str, Path]) -> None:  # type: ignore
+    def export(self, filename: Union[str, Path]) -> None:  # type: ignore
         """ Export to disk if a different location
 
             Args:
@@ -345,17 +375,18 @@ class BloomFilterOnDisk(BaseBloom):
             Note:
                 Only exported if the filename is not the original filename """
         self.__update()
-        filename = Path(filename)
-        if filename.name != self.__filename.name:
+        if filename and Path(filename).exists():
+            filename = Path(filename).name
+        if filename != self.__filename:
             # setup the new bloom filter
-            copyfile(self.__filename.name, filename.name)
+            copyfile(self.__filename.name, str(filename))
         # otherwise, nothing to do!
 
     def add_alt(self, hashes: HashResultsT) -> None:
         super(BloomFilterOnDisk, self).add_alt(hashes)
         self.__update()
 
-    def union(self, second: SimpleBloomT) -> typing.Optional["BaseBloom"]:
+    def union(self, second: SimpleBloomT) -> Union["BaseBloom", None]:
         """ Return a new Bloom Filter that contains the union of the two
 
             Args:
@@ -379,7 +410,7 @@ class BloomFilterOnDisk(BaseBloom):
 
         return _tmp_union(self, second)
 
-    def intersection(self, second: SimpleBloomT) -> typing.Optional["BaseBloom"]:
+    def intersection(self, second: SimpleBloomT) -> Union["BaseBloom", None]:
         """ Return a new Bloom Filter that contains the intersection of the
             two
 
@@ -404,7 +435,7 @@ class BloomFilterOnDisk(BaseBloom):
 
         return _tmp_intersection(self, second)
 
-    def jaccard_index(self, second: SimpleBloomT) -> typing.Optional[float]:
+    def jaccard_index(self, second: SimpleBloomT) -> Union[float, None]:
         """ Calculate the jaccard similarity score between two Bloom Filters
 
             Args:
@@ -427,21 +458,21 @@ class BloomFilterOnDisk(BaseBloom):
             return None
         return _tmp_jaccard_index(self, second)
 
-    def _load_hex(self, hex_string: str, hash_function: typing.Optional[HashFuncT] = None):
+    def _load_hex(self, hex_string: str, hash_function: Union[HashFuncT, None] = None):
         """load from hex ..."""
         msg = "Loading from hex_string is currently not supported by the on disk Bloom Filter"
         raise NotSupportedError(msg)
 
     def _get_element(self, idx: int) -> int:
         """wrappper to use similar functions always!"""
-        return unpack("B", bytes([self._bloom[idx]]))[0]  # type: ignore
+        return int(self.CNT_ELM_STRUCT.unpack(bytes([self._bloom[idx]]))[0])
 
     def __update(self):
         """update the on disk Bloom Filter and ensure everything is out
         to disk"""
         self._bloom.flush()
-        self.__file_pointer.seek(-self.__export_offset, os.SEEK_END)
-        self.__file_pointer.write(pack("Q", self.elements_added))
+        self.__file_pointer.seek(-self.CNT_EXPORT_OFFSET.size, os.SEEK_END)
+        self.__file_pointer.write(self.CNT_EXP_ELM_STRUCT.pack(self.elements_added))
         self.__file_pointer.flush()
 
     def _cnt_number_bits_set(self) -> int:
