@@ -4,7 +4,10 @@
     URL: https://github.com/barrust/counting_bloom
 """
 
-import typing
+from collections.abc import ByteString
+from pathlib import Path
+from struct import Struct
+from typing import Union
 
 from ..constants import UINT32_T_MAX, UINT64_T_MAX
 from ..hashes import HashFuncT, HashResultsT, KeyT
@@ -45,11 +48,11 @@ class CountingBloomFilter(BaseBloom):
 
     def __init__(
         self,
-        est_elements: typing.Optional[int] = None,
-        false_positive_rate: typing.Optional[float] = None,
-        filepath: typing.Optional[str] = None,
-        hex_string: typing.Optional[str] = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        est_elements: Union[int, None] = None,
+        false_positive_rate: Union[float, None] = None,
+        filepath: Union[str, Path, None] = None,
+        hex_string: Union[str, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
         """setup the basic values needed"""
         super(CountingBloomFilter, self).__init__(
@@ -61,6 +64,27 @@ class CountingBloomFilter(BaseBloom):
             hash_function=hash_function,
         )
 
+    __HEADER_STRUCT_FORMAT = "QQf"
+    __HEADER_STRUCT = Struct(__HEADER_STRUCT_FORMAT)
+    # __HEADER_STRUCT_BE = Struct(">" + __HEADER_STRUCT_FORMAT)
+
+    @classmethod
+    def frombytes(cls, b: ByteString, hash_function: Union[HashFuncT, None] = None) -> "CountingBloomFilter":
+        """
+        Args:
+            b (ByteString): the bytes to load as a Counting Bloom Filter
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            CountingBloomFilter: A Counting Bloom Filter object
+        """
+        offset = cls.__HEADER_STRUCT.size
+        est_elms, _, fpr, _, _, _ = cls._parse_footer(cls.__HEADER_STRUCT, bytes(b[-offset:]))
+        blm = CountingBloomFilter(est_elements=est_elms, false_positive_rate=fpr, hash_function=hash_function)
+        blm._parse_footer_set(cls.__HEADER_STRUCT, bytes(b[-offset:]))
+        blm._set_bloom_length()
+        blm._parse_bloom_array(b)
+        return blm
+
     def __str__(self) -> str:
         """string representation of the counting bloom filter"""
         on_disk = "no" if self.is_on_disk is False else "yes"
@@ -246,7 +270,7 @@ class CountingBloomFilter(BaseBloom):
         res.elements_added = res.estimate_elements()
         return res
 
-    def jaccard_index(self, second: "CountingBloomFilter") -> typing.Optional[float]:
+    def jaccard_index(self, second: "CountingBloomFilter") -> Union[float, None]:
         """ Take the Jaccard Index of two Counting Bloom Filters
 
             Args:
@@ -280,7 +304,7 @@ class CountingBloomFilter(BaseBloom):
             return 1.0
         return count_inter / count_union
 
-    def union(self, second: "CountingBloomFilter") -> typing.Optional["CountingBloomFilter"]:
+    def union(self, second: "CountingBloomFilter") -> Union["CountingBloomFilter", None]:
         """ Return a new Countiong Bloom Filter that contains the union of
             the two
 
