@@ -5,13 +5,14 @@
 """
 
 import math
-import os
-import typing
+from array import array
+from collections.abc import ByteString
 from io import BytesIO, IOBase
 from mmap import mmap
 from numbers import Number
 from pathlib import Path
-from struct import calcsize, pack, unpack
+from struct import Struct
+from typing import Dict, Tuple, Union
 
 from ..constants import INT32_T_MAX, INT32_T_MIN, INT64_T_MAX, INT64_T_MIN
 from ..exceptions import CountMinSketchError, InitializationError, NotSupportedError
@@ -59,12 +60,12 @@ class CountMinSketch(object):
 
     def __init__(
         self,
-        width: typing.Optional[int] = None,
-        depth: typing.Optional[int] = None,
-        confidence: typing.Optional[float] = None,
-        error_rate: typing.Optional[float] = None,
-        filepath: typing.Optional[typing.Union[str, Path]] = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        width: Union[int, None] = None,
+        depth: Union[int, None] = None,
+        confidence: Union[float, None] = None,
+        error_rate: Union[float, None] = None,
+        filepath: Union[str, Path, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
         """default initilization function"""
         # default values
@@ -77,44 +78,51 @@ class CountMinSketch(object):
 
         if is_valid_file(filepath):
             self.__load(filepath)
-        elif width is not None and depth is not None:
-            valid_prms = isinstance(width, Number) and width > 0 and isinstance(depth, Number) and depth > 0
-            if not valid_prms:
-                msg = "CountMinSketch: width and depth must be greater than 0"
-                raise InitializationError(msg)
-            self.__width = int(width)
-            self.__depth = int(depth)
-            self.__confidence = 1 - (1 / math.pow(2, self.depth))
-            self.__error_rate = 2 / self.width
-            self._bins = [0] * (self.width * self.depth)
-        elif confidence is not None and error_rate is not None:
-            valid_prms = (
-                isinstance(confidence, Number) and confidence > 0 and isinstance(error_rate, Number) and error_rate > 0
-            )
-            if not valid_prms:
-                msg = "CountMinSketch: width and depth must be greater than 0"
-                raise InitializationError(msg)
-            self.__confidence = confidence
-            self.__error_rate = error_rate
-            self.__width = math.ceil(2 / error_rate)
-            numerator = -1 * math.log(1 - confidence)
-            self.__depth = math.ceil(numerator / 0.6931471805599453)
-            self._bins = [0] * int(self.width * self.depth)
         else:
-            msg = (
-                "Must provide one of the following to initialize the "
-                "Count-Min Sketch:\n"
-                "    A file to load,\n"
-                "    The width and depth,\n"
-                "    OR confidence and error rate"
-            )
-            raise InitializationError(msg)
+            if width is not None and depth is not None:
+                valid_prms = isinstance(width, Number) and width > 0 and isinstance(depth, Number) and depth > 0
+                if not valid_prms:
+                    msg = "CountMinSketch: width and depth must be greater than 0"
+                    raise InitializationError(msg)
+                self.__width = int(width)
+                self.__depth = int(depth)
+                self.__confidence = 1 - (1 / math.pow(2, self.depth))
+                self.__error_rate = 2 / self.width
+            elif confidence is not None and error_rate is not None:
+                valid_prms = (
+                    isinstance(confidence, Number)
+                    and confidence > 0
+                    and isinstance(error_rate, Number)
+                    and error_rate > 0
+                )
+                if not valid_prms:
+                    msg = "CountMinSketch: width and depth must be greater than 0"
+                    raise InitializationError(msg)
+                self.__confidence = confidence
+                self.__error_rate = error_rate
+                self.__width = math.ceil(2 / error_rate)
+                numerator = -1 * math.log(1 - confidence)
+                self.__depth = math.ceil(numerator / 0.6931471805599453)
+
+            else:
+                msg = (
+                    "Must provide one of the following to initialize the "
+                    "Count-Min Sketch:\n"
+                    "    A file to load,\n"
+                    "    The width and depth,\n"
+                    "    OR confidence and error rate"
+                )
+                raise InitializationError(msg)
+            self._bins = array("i", [0]) * (self.width * self.depth)
 
         if hash_function is None:
             self._hash_function = default_fnv_1a
         else:
             self._hash_function = hash_function  # type: ignore
 
+    __FOOTER_STRUCT = Struct("IIq")
+    __BASIC_BIN_STRUCT = Struct("i")
+
     def __str__(self) -> str:
         """string representation of the count min sketch"""
         msg = (
@@ -143,6 +151,21 @@ class CountMinSketch(object):
             self.export(f)
             return f.getvalue()
 
+    @classmethod
+    def frombytes(cls, b: ByteString, hash_function: Union[HashFuncT, None] = None) -> "CountMinSketch":
+        """
+        Args:
+            b (ByteString): The bytes to load as a Count-Min Sketch
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            CountMinSketch: A count-min sketch object
+        """
+        offset = cls.__FOOTER_STRUCT.size
+        width, depth, _ = cls.__FOOTER_STRUCT.unpack_from(bytes(b[-offset:]))
+        cms = CountMinSketch(width=width, depth=depth, hash_function=hash_function)
+        cms._parse_bytes(b)
+        return cms
+
     @property
     def width(self) -> int:
         """int: The width of the count-min sketch
@@ -221,7 +244,7 @@ class CountMinSketch(object):
         for i, _ in enumerate(self._bins):
             self._bins[i] = 0
 
-    def hashes(self, key: KeyT, depth: typing.Optional[int] = None) -> HashResultsT:
+    def hashes(self, key: KeyT, depth: Union[int, None] = None) -> HashResultsT:
         """ Return the hashes based on the provided key
 
             Args:
@@ -258,9 +281,12 @@ class CountMinSketch(object):
         res = list()
         for i, val in enumerate(hashes):
             t_bin = (val % self.width) + (i * self.width)
-            self._bins[t_bin] += num_els
-            if self._bins[t_bin] > INT32_T_MAX:
+            tmp_els = self._bins[t_bin] + num_els
+
+            if tmp_els > INT32_T_MAX:
                 self._bins[t_bin] = INT32_T_MAX
+            else:
+                self._bins[t_bin] = tmp_els
             res.append(self._bins[t_bin])
         self.__elements_added += num_els
 
@@ -292,9 +318,11 @@ class CountMinSketch(object):
         res = list()
         for i, val in enumerate(hashes):
             t_bin = (val % self.width) + (i * self.width)
-            self._bins[t_bin] -= num_els
-            if self._bins[t_bin] < INT32_T_MIN:
+            tmp_els = self._bins[t_bin] - num_els
+            if tmp_els < INT32_T_MIN:
                 self._bins[t_bin] = INT32_T_MIN
+            else:
+                self._bins[t_bin] = tmp_els
             res.append(self._bins[t_bin])
         self.__elements_added -= num_els
         if self.elements_added < INT64_T_MIN:
@@ -323,7 +351,7 @@ class CountMinSketch(object):
         bins = self.__get_values_sorted(hashes)
         return self.__query_method(bins)
 
-    def export(self, file: typing.Union[Path, str, IOBase, mmap]) -> None:
+    def export(self, file: Union[Path, str, IOBase, mmap]) -> None:
         """ Export the count-min sketch to disk
 
             Args:
@@ -334,9 +362,8 @@ class CountMinSketch(object):
                 self.export(filepointer)  # type: ignore
         else:
             # write out the bins
-            rep = "i" * len(self._bins)
-            file.write(pack(rep, *self._bins))
-            file.write(pack("IIq", self.width, self.depth, self.elements_added))
+            self._bins.tofile(file)  # type: ignore
+            file.write(self.__FOOTER_STRUCT.pack(self.width, self.depth, self.elements_added))
 
     def join(self, second: "CountMinSketch") -> None:
         """ Join two count-min sketchs into a single count-min sketch; the
@@ -367,11 +394,14 @@ class CountMinSketch(object):
         for i in range(size):
             if self._bins[i] == INT32_T_MIN or self._bins[i] == INT32_T_MAX:
                 continue
-            self._bins[i] += second._bins[i]
-            if self._bins[i] > INT32_T_MAX:
+            tmp_els = self._bins[i] + second._bins[i]
+
+            if tmp_els > INT32_T_MAX:
                 self._bins[i] = INT32_T_MAX
-            elif self._bins[i] < INT32_T_MIN:
+            elif tmp_els < INT32_T_MIN:
                 self._bins[i] = INT32_T_MIN
+            else:
+                self._bins[i] = tmp_els
 
         # handle adding and removing elements added including handling overflow
         self.__elements_added += self.elements_added
@@ -381,27 +411,33 @@ class CountMinSketch(object):
         elif self.elements_added < INT64_T_MIN:
             self.__elements_added = INT64_T_MIN
 
-    def __load(self, file: typing.Union[Path, str, IOBase, mmap]):
+    def __load(self, file: Union[Path, str, IOBase, mmap]):
         """load the count-min sketch from file"""
         if not isinstance(file, (IOBase, mmap)):
             file = Path(file)
             with MMap(file) as filepointer:
                 self.__load(filepointer)
         else:
-            offset = calcsize("IIq")
-            file.seek(offset * -1, os.SEEK_END)
-            mybytes = unpack("IIq", file.read(offset))
-            self.__width = mybytes[0]
-            self.__depth = mybytes[1]
-            self.__elements_added = mybytes[2]
-            self.__confidence = 1 - (1 / math.pow(2, self.depth))
-            self.__error_rate = 2 / self.width
+            self._parse_bytes(file)  # type: ignore
 
-            file.seek(0, os.SEEK_SET)
-            length = self.width * self.depth
-            rep = "i" * length
-            offset = calcsize(rep)
-            self._bins = list(unpack(rep, file.read(offset)))
+    @classmethod
+    def _parse_footer(cls, file: ByteString) -> Tuple[int, int, int]:
+        """return width, depth and elements added, in that order"""
+        offset = cls.__FOOTER_STRUCT.size
+        width, depth, elements_added = cls.__FOOTER_STRUCT.unpack_from(bytes(file[-offset:]))
+        return width, depth, elements_added
+
+    def _parse_bytes(self, file: ByteString):
+        """parse bytes or a mapped file to setup the CountMin-Sketch"""
+        width, depth, els_added = self._parse_footer(file)
+        self.__width = width
+        self.__depth = depth
+        self.__elements_added = els_added
+        self.__confidence = 1 - (1 / math.pow(2, self.depth))
+        self.__error_rate = 2 / self.width
+
+        offset = self.__BASIC_BIN_STRUCT.size * self.width * self.depth
+        self._bins = array("i", bytes(file[:offset]))
 
     def __get_values_sorted(self, hashes: HashResultsT) -> HashResultsT:
         """get the values sorted"""
@@ -469,12 +505,12 @@ class CountMeanSketch(CountMinSketch):
 
     def __init__(
         self,
-        width: typing.Optional[int] = None,
-        depth: typing.Optional[int] = None,
-        confidence: typing.Optional[float] = None,
-        error_rate: typing.Optional[float] = None,
-        filepath: str = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        width: Union[int, None] = None,
+        depth: Union[int, None] = None,
+        confidence: Union[float, None] = None,
+        error_rate: Union[float, None] = None,
+        filepath: Union[str, Path, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
         super(CountMeanSketch, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
         self.query_type = "mean"
@@ -510,12 +546,12 @@ class CountMeanMinSketch(CountMinSketch):
 
     def __init__(
         self,
-        width: typing.Optional[int] = None,
-        depth: typing.Optional[int] = None,
-        confidence: typing.Optional[float] = None,
-        error_rate: typing.Optional[float] = None,
-        filepath: typing.Optional[str] = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        width: Union[int, None] = None,
+        depth: Union[int, None] = None,
+        confidence: Union[float, None] = None,
+        error_rate: Union[float, None] = None,
+        filepath: Union[str, Path, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
         super(CountMeanMinSketch, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
         self.query_type = "mean-min"
@@ -552,12 +588,12 @@ class HeavyHitters(CountMinSketch):
     def __init__(
         self,
         num_hitters: int = 100,
-        width: typing.Optional[int] = None,
-        depth: typing.Optional[int] = None,
-        confidence: typing.Optional[float] = None,
-        error_rate: typing.Optional[float] = None,
-        filepath: typing.Optional[typing.Union[str, Path]] = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        width: Union[int, None] = None,
+        depth: Union[int, None] = None,
+        confidence: Union[float, None] = None,
+        error_rate: Union[float, None] = None,
+        filepath: Union[str, Path, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
 
         super(HeavyHitters, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
@@ -566,6 +602,23 @@ class HeavyHitters(CountMinSketch):
         self.__num_hitters = num_hitters
         self.__smallest = 0
 
+    @classmethod
+    def frombytes(  # type: ignore
+        cls, b: ByteString, num_hitters: int = 100, hash_function: Union[HashFuncT, None] = None
+    ) -> "HeavyHitters":
+        """
+        Args:
+            b (ByteString): The bytes to load as a Expanding Bloom Filter
+            num_hitters (int): The maximum number of distinct elements to track
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            HeavyHitters: A Bloom Filter object
+        """
+        width, depth, _ = cls._parse_footer(b)
+        hh = HeavyHitters(width=width, depth=depth, num_hitters=num_hitters, hash_function=hash_function)
+        hh._parse_bytes(b)
+        return hh
+
     def __str__(self) -> str:
         """heavy hitters string rep"""
         msg = super(HeavyHitters, self).__str__()
@@ -573,7 +626,7 @@ class HeavyHitters(CountMinSketch):
         return tmp.format(msg, self.number_heavy_hitters, self.__top_x_size)
 
     @property
-    def heavy_hitters(self) -> typing.Dict[str, int]:
+    def heavy_hitters(self) -> Dict[str, int]:
         """dict: Return the heavy hitters, or most common elements
 
         Note:
@@ -670,24 +723,62 @@ class HeavyHitters(CountMinSketch):
 
 
 class StreamThreshold(CountMinSketch):
-    """keep track of those elements over a certain threshold"""
+    """ Find and track those elements that are above a certain threshold
+
+        Args:
+            threshold (int): The threshold above which an element will be tracked
+            width (int): The width of the count-min sketch
+            depth (int): The depth of the count-min sketch
+            confidence (float): The level of confidence desired
+            error_rate (float): The desired error rate
+            filepath (str): Path to file to load
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            StreamThreshold: A Count-Min Sketch object
+        Note:
+            Initialization order of operations:
+                1) From file
+                2) Width and depth
+                3) Confidence and error rate
+        Note:
+            Default query type is `min`
+        Note:
+            For width and depth, width may realistically be in the thousands \
+            while depth is in the single digit to teens  """
 
     __slots__ = ["__threshold", "__meets_threshold"]
 
     def __init__(
         self,
         threshold: int = 100,
-        width: typing.Optional[int] = None,
-        depth: typing.Optional[int] = None,
-        confidence: typing.Optional[float] = None,
-        error_rate: typing.Optional[float] = None,
-        filepath: typing.Optional[str] = None,
-        hash_function: typing.Optional[HashFuncT] = None,
+        width: Union[int, None] = None,
+        depth: Union[int, None] = None,
+        confidence: Union[float, None] = None,
+        error_rate: Union[float, None] = None,
+        filepath: Union[str, Path, None] = None,
+        hash_function: Union[HashFuncT, None] = None,
     ) -> None:
         super(StreamThreshold, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
         self.__threshold = threshold
         self.__meets_threshold = dict()  # type: ignore
 
+    @classmethod
+    def frombytes(  # type: ignore
+        cls, b: ByteString, threshold: int = 100, hash_function: Union[HashFuncT, None] = None
+    ) -> "StreamThreshold":
+        """
+        Args:
+            b (ByteString): The bytes to load as a Expanding Bloom Filter
+            threshold (int): The threshold above which an element will be tracked
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            StreamThreshold: A Bloom Filter object
+        """
+        width, depth, _ = cls._parse_footer(b)
+        st = StreamThreshold(width=width, depth=depth, threshold=threshold, hash_function=hash_function)
+        st._parse_bytes(b)
+        return st
+
     def __str__(self) -> str:
         """stream threshold string rep"""
         msg = super(StreamThreshold, self).__str__()
@@ -695,7 +786,7 @@ class StreamThreshold(CountMinSketch):
         return tmp.format(msg, self.threshold, len(self.__meets_threshold))
 
     @property
-    def meets_threshold(self) -> typing.Dict[str, int]:
+    def meets_threshold(self) -> Dict[str, int]:
         """dict: Those keys that meet the required threshold (with value)"""
         return self.__meets_threshold
 
