@@ -3,13 +3,14 @@
     Author: Tyler Barrus (barrust@gmail.com)
 """
 
-import os
 import random
-import typing
+from array import array
+from collections.abc import ByteString
 from io import IOBase
 from mmap import mmap
 from pathlib import Path
-from struct import calcsize, pack, unpack
+from struct import Struct
+from typing import List, Union
 
 from ..exceptions import CuckooFilterFullError
 from ..hashes import KeyT, SimpleHashT
@@ -40,6 +41,7 @@ class CountingCuckooFilter(CuckooFilter):
         "__max_cuckoo_swaps",
         "_cuckoo_capacity",
         "_buckets",
+        "_fingerprint_size",
     ]
 
     def __init__(
@@ -50,8 +52,8 @@ class CountingCuckooFilter(CuckooFilter):
         expansion_rate: int = 2,
         auto_expand: bool = True,
         finger_size: int = 4,
-        filepath: typing.Optional[str] = None,
-        hash_function: typing.Optional[SimpleHashT] = None,  # this is INCORRECT!
+        filepath: Union[str, Path, None] = None,
+        hash_function: Union[SimpleHashT, None] = None,
     ) -> None:
         """setup the data structure"""
         self.__unique_elements = 0
@@ -66,6 +68,81 @@ class CountingCuckooFilter(CuckooFilter):
             hash_function,
         )
 
+    __FOOTER_STRUCT = Struct("II")
+    __BIN_STRUCT = Struct("II")
+
+    @classmethod
+    def init_error_rate(
+        cls,
+        error_rate: float,
+        capacity: int = 10000,
+        bucket_size: int = 4,
+        max_swaps: int = 500,
+        expansion_rate: int = 2,
+        auto_expand: bool = True,
+        hash_function: Union[SimpleHashT, None] = None,
+    ):
+        """Initialize a simple Cuckoo Filter based on error rate
+
+        Args:
+            error_rate (float):
+            capacity (int): The number of bins
+            bucket_size (int): The number of buckets per bin
+            max_swaps (int): The number of cuckoo swaps before stopping
+            expansion_rate (int): The rate at which to expand
+            auto_expand (bool): If the filter should automatically expand
+            hash_function (function): Hashing strategy function to use \
+            `hf(key)`
+        Returns:
+            CuckooFilter: A Cuckoo Filter object"""
+        cku = CountingCuckooFilter(
+            capacity=capacity,
+            bucket_size=bucket_size,
+            auto_expand=auto_expand,
+            max_swaps=max_swaps,
+            expansion_rate=expansion_rate,
+            hash_function=hash_function,
+        )
+        cku._set_error_rate(error_rate)
+        return cku
+
+    @classmethod
+    def load_error_rate(
+        cls, error_rate: float, filepath: Union[str, Path], hash_function: Union[SimpleHashT, None] = None
+    ):
+        """Initialize a previously exported Cuckoo Filter based on error rate
+
+        Args:
+            error_rate (float):
+            filepath (str): The path to the file to load or None if no file
+            hash_function (function): Hashing strategy function to use \
+            `hf(key)`
+        Returns:
+            CuckooFilter: A Cuckoo Filter object
+        """
+        cku = CountingCuckooFilter(filepath=filepath, hash_function=hash_function)
+        cku._set_error_rate(error_rate)
+        return cku
+
+    @classmethod
+    def frombytes(
+        cls, b: ByteString, error_rate: Union[float, None] = None, hash_function: Union[SimpleHashT, None] = None
+    ) -> "CountingCuckooFilter":
+        """
+        Args:
+            b (ByteString): The bytes to load as a Expanding Bloom Filter
+            error_rate (float): The error rate of the cuckoo filter, if used to generate the original filter
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            CountingCuckooFilter: A Bloom Filter object
+        """
+        cku = CountingCuckooFilter(hash_function=hash_function)
+        cku._load(b)
+
+        # if error rate is provided, use it
+        cku._set_error_rate(error_rate)
+        return cku
+
     def __contains__(self, val: KeyT) -> bool:
         """setup the `in` keyword"""
         if self.check(val) > 0:
@@ -78,7 +155,7 @@ class CountingCuckooFilter(CuckooFilter):
         return self.__unique_elements
 
     @property
-    def buckets(self) -> typing.List[typing.List["CountingCuckooBin"]]:  # type: ignore
+    def buckets(self) -> List[List["CountingCuckooBin"]]:  # type: ignore
         """list(list): The buckets holding the fingerprints
 
         Note:
@@ -150,7 +227,7 @@ class CountingCuckooFilter(CuckooFilter):
         """Expand the cuckoo filter"""
         self._expand_logic(None)
 
-    def export(self, file: typing.Union[Path, str, IOBase, mmap]) -> None:
+    def export(self, file: Union[Path, str, IOBase, mmap]) -> None:
         """Export cuckoo filter to file
 
         Args:
@@ -161,19 +238,15 @@ class CountingCuckooFilter(CuckooFilter):
         else:
             filepointer = file  # type:ignore
             for bucket in self.buckets:
-                # do something for each...
-                rep = len(bucket) * "II"
-                wbyt = pack(rep, *[x for x in self.__bucket_decomposition(bucket)])
-                filepointer.write(wbyt)
+                self.__bucket_decomposition(bucket).tofile(filepointer)
                 leftover = self.bucket_size - len(bucket)
-                rep = leftover * "II"
-                filepointer.write(pack(rep, *([0] * (leftover * 2))))
+                array("I", [0 for _ in range(leftover * 2)]).tofile(filepointer)
             # now put out the required information at the end
-            filepointer.write(pack("II", self.bucket_size, self.max_swaps))
+            filepointer.write(self.__FOOTER_STRUCT.pack(self.bucket_size, self.max_swaps))
 
     def _insert_fingerprint_alt(
         self, fingerprint: int, idx_1: int, idx_2: int, count: int = 1
-    ) -> typing.Optional["CountingCuckooBin"]:
+    ) -> Union["CountingCuckooBin", None]:
         """insert a fingerprint, but with a count parameter!"""
         if self.__insert_element(fingerprint, idx_1, count):
             self._inserted_elements += 1
@@ -208,7 +281,7 @@ class CountingCuckooFilter(CuckooFilter):
         # if we got here we have an error... we might need to know what is left
         return prv_bin
 
-    def _check_if_present(self, idx_1: int, idx_2: int, fingerprint: int) -> typing.Optional[int]:
+    def _check_if_present(self, idx_1: int, idx_2: int, fingerprint: int) -> Union[int, None]:
         """wrapper for checking if fingerprint is already inserted"""
         if fingerprint in [x.finger for x in self.buckets[idx_1]]:
             return idx_1
@@ -216,34 +289,39 @@ class CountingCuckooFilter(CuckooFilter):
             return idx_2
         return None
 
-    def _load(self, file: typing.Union[Path, str, IOBase, mmap]) -> None:
+    def _load(self, file: Union[Path, str, IOBase, mmap, bytes, ByteString]) -> None:
         """load a cuckoo filter from file"""
-        if not isinstance(file, (IOBase, mmap)):
+        if not isinstance(file, (IOBase, mmap, bytes, ByteString)):
             file = Path(file)
             with MMap(file) as filepointer:
                 self._load(filepointer)
         else:
-            offset = calcsize("II")
-            int_size = calcsize("II")
-            file.seek(offset * -1, os.SEEK_END)
-            list_size = file.tell()
-            mybytes = unpack("II", file.read(offset))
-            self._bucket_size = mybytes[0]
-            self.__max_cuckoo_swaps = mybytes[1]
-            self._cuckoo_capacity = list_size // int_size // self.bucket_size
+            self._parse_footer(file)  # type: ignore
             self._inserted_elements = 0
-            # now pull everything in!
-            file.seek(0, os.SEEK_SET)
-            self._buckets = list()
-            for i in range(self.capacity):
-                self.buckets.append(list())
-                for _ in range(self.bucket_size):
-                    finger, count = unpack("II", file.read(int_size))
-                    if finger > 0:
-                        ccb = CountingCuckooBin(finger, count)
-                        self.buckets[i].append(ccb)
-                        self._inserted_elements += count
-                        self.__unique_elements += 1
+            self._parse_buckets(file)  # type: ignore
+
+    def _parse_footer(self, d: ByteString) -> None:
+        bucket_size, max_swaps = self.__FOOTER_STRUCT.unpack(bytes(d[-self.__FOOTER_STRUCT.size :]))
+        self._bucket_size = int(bucket_size)
+        self.__max_cuckoo_swaps = int(max_swaps)
+
+    def _parse_buckets(self, d: ByteString) -> None:
+        bin_size = self.__BIN_STRUCT.size
+        self._cuckoo_capacity = (len(bytes(d)) - bin_size) // bin_size // self.bucket_size
+        start = 0
+        end = bin_size
+        self._buckets = list()
+        for i in range(self.capacity):
+            self.buckets.append(list())
+            for _ in range(self.bucket_size):
+                finger, count = self.__BIN_STRUCT.unpack(bytes(d[start:end]))
+                if finger > 0:
+                    ccb = CountingCuckooBin(finger, count)
+                    self.buckets[i].append(ccb)
+                    self._inserted_elements += count
+                    self.__unique_elements += 1
+                start = end
+                end += bin_size
 
     def _expand_logic(self, extra_fingerprint: "CountingCuckooBin") -> None:
         """the logic to acutally expand the cuckoo filter"""
@@ -266,36 +344,40 @@ class CountingCuckooFilter(CuckooFilter):
         return False
 
     @staticmethod
-    def __bucket_decomposition(bucket):
+    def __bucket_decomposition(bucket: List["CountingCuckooBin"]):
+        arr = array("I")
         for buck in bucket:
-            yield buck.finger
-            yield buck.count
+            arr.extend(buck.get_array())
+        return arr
 
 
 class CountingCuckooBin(object):
     """A container class for the counting cuckoo filter"""
 
     # keep it lightweight
-    __slots__ = ["__fingerprint", "__count"]
+    __slots__ = ["__bin"]
 
     def __init__(self, fingerprint: int, count: int) -> None:
         """init"""
-        self.__fingerprint = fingerprint
-        self.__count = count
+        self.__bin = array("I", [fingerprint, count])
 
     def __contains__(self, val: int) -> bool:
         """setup the `in` construct"""
-        return self.__fingerprint == val
+        return self.__bin[0] == val
+
+    def get_array(self):
+        """return the array implementation"""
+        return self.__bin
 
     @property
     def finger(self) -> int:
         """fingerprint property"""
-        return self.__fingerprint
+        return self.__bin[0]
 
     @property
     def count(self) -> int:
         """count property"""
-        return self.__count
+        return self.__bin[1]
 
     def __repr__(self) -> str:
         """how do we represent this?"""
@@ -303,14 +385,14 @@ class CountingCuckooBin(object):
 
     def __str__(self) -> str:
         """convert it into a string"""
-        return "(fingerprint:{} count:{})".format(self.__fingerprint, self.__count)
+        return "(fingerprint:{} count:{})".format(self.__bin[0], self.__bin[1])
 
     def increment(self) -> int:
         """increment"""
-        self.__count += 1
-        return self.__count
+        self.__bin[1] += 1
+        return self.__bin[1]
 
     def decrement(self) -> int:
         """decrement"""
-        self.__count -= 1
-        return self.__count
+        self.__bin[1] -= 1
+        return self.__bin[1]
