@@ -3,20 +3,20 @@
     Author: Tyler Barrus (barrust@gmail.com)
 """
 
-import array
-import os
+import math
 import random
-import typing
+from array import array
 from collections.abc import ByteString
 from io import BytesIO, IOBase
 from mmap import mmap
 from numbers import Number
 from pathlib import Path
-from struct import Struct, calcsize, pack
+from struct import Struct
+from typing import List, Tuple, Union
 
 from ..exceptions import CuckooFilterFullError, InitializationError
 from ..hashes import KeyT, SimpleHashT, fnv_1a
-from ..utilities import MMap, convert_to_typed, get_x_bits, is_valid_file
+from ..utilities import MMap, get_x_bits, is_valid_file
 
 
 class CuckooFilter(object):
@@ -31,7 +31,7 @@ class CuckooFilter(object):
             finger_size (int): The size of the fingerprint to use in bytes \
             (between 1 and 4); exported as 4 bytes; up to the user to reset \
             the size correctly on import
-            filename (str): The path to the file to load or None if no file
+            filepath (str): The path to the file to load or None if no file
             hash_function (function): Hashing strategy function to use \
             `hf(key)`
         Returns:
@@ -43,10 +43,11 @@ class CuckooFilter(object):
         "__max_cuckoo_swaps",
         "__expansion_rate",
         "__auto_expand",
-        "__fingerprint_size",
+        "_fingerprint_size",
         "__hash_func",
         "_inserted_elements",
         "_buckets",
+        "_error_rate",
     ]
 
     def __init__(
@@ -57,8 +58,8 @@ class CuckooFilter(object):
         expansion_rate: int = 2,
         auto_expand: bool = True,
         finger_size: int = 4,
-        filepath: typing.Optional[str] = None,
-        hash_function: typing.Optional[SimpleHashT] = None,  # this is incorrect
+        filepath: Union[str, Path, None] = None,
+        hash_function: Union[SimpleHashT, None] = None,
     ):
         """setup the data structure"""
         valid_prms = (
@@ -75,11 +76,11 @@ class CuckooFilter(object):
         self._bucket_size = int(bucket_size)
         self._cuckoo_capacity = int(capacity)
         self.__max_cuckoo_swaps = int(max_swaps)
-        self.__expansion_rate = None
+        self.__expansion_rate = 2
         self.expansion_rate = expansion_rate
         self.__auto_expand = True
         self.auto_expand = auto_expand
-        self.__fingerprint_size = None
+        self._fingerprint_size = 32
         self.fingerprint_size = finger_size
 
         if hash_function is None:
@@ -97,6 +98,80 @@ class CuckooFilter(object):
             msg = "CuckooFilter: failed to load provided file"
             raise InitializationError(msg)
 
+        self._error_rate = float(self._calc_error_rate())
+
+    @classmethod
+    def init_error_rate(
+        cls,
+        error_rate: float,
+        capacity: int = 10000,
+        bucket_size: int = 4,
+        max_swaps: int = 500,
+        expansion_rate: int = 2,
+        auto_expand: bool = True,
+        hash_function: Union[SimpleHashT, None] = None,
+    ):
+        """Initialize a simple Cuckoo Filter based on error rate
+
+        Args:
+            error_rate (float):
+            capacity (int): The number of bins
+            bucket_size (int): The number of buckets per bin
+            max_swaps (int): The number of cuckoo swaps before stopping
+            expansion_rate (int): The rate at which to expand
+            auto_expand (bool): If the filter should automatically expand
+            hash_function (function): Hashing strategy function to use \
+            `hf(key)`
+        Returns:
+            CuckooFilter: A Cuckoo Filter object"""
+        cku = CuckooFilter(
+            capacity=capacity,
+            bucket_size=bucket_size,
+            auto_expand=auto_expand,
+            max_swaps=max_swaps,
+            expansion_rate=expansion_rate,
+            hash_function=hash_function,
+        )
+        cku._set_error_rate(error_rate)
+        return cku
+
+    @classmethod
+    def load_error_rate(
+        cls, error_rate: float, filepath: Union[str, Path], hash_function: Union[SimpleHashT, None] = None
+    ):
+        """Initialize a previously exported Cuckoo Filter based on error rate
+
+        Args:
+            error_rate (float):
+            filepath (str): The path to the file to load or None if no file
+            hash_function (function): Hashing strategy function to use \
+            `hf(key)`
+        Returns:
+            CuckooFilter: A Cuckoo Filter object
+        """
+        cku = CuckooFilter(filepath=filepath, hash_function=hash_function)
+        cku._set_error_rate(error_rate)
+        return cku
+
+    @classmethod
+    def frombytes(
+        cls, b: ByteString, error_rate: Union[float, None] = None, hash_function: Union[SimpleHashT, None] = None
+    ) -> "CuckooFilter":
+        """
+        Args:
+            b (ByteString): The bytes to load as a Expanding Bloom Filter
+            error_rate (float): The error rate of the cuckoo filter, if used to generate the original filter
+            hash_function (function): Hashing strategy function to use `hf(key, number)`
+        Returns:
+            CuckooFilter: A Bloom Filter object
+        """
+        cku = CuckooFilter(hash_function=hash_function)
+        cku._load(b)  # type: ignore
+
+        # if error rate is provided, use it
+        cku._set_error_rate(error_rate)
+        return cku
+
     def __contains__(self, key: KeyT) -> bool:
         """setup the `in` keyword"""
         return self.check(key)
@@ -157,7 +232,7 @@ class CuckooFilter(object):
         return self._bucket_size
 
     @property
-    def buckets(self) -> typing.List[typing.List[int]]:  # TODO: not sure this is correct
+    def buckets(self) -> List[List[int]]:
         """list(list): The buckets holding the fingerprints
 
         Note:
@@ -174,6 +249,11 @@ class CuckooFilter(object):
         """set the self expand value"""
         self.__expansion_rate = int(val)
 
+    @property
+    def error_rate(self) -> float:
+        """float: The error rate of the cuckoo filter"""
+        return self._error_rate
+
     @property
     def auto_expand(self) -> bool:
         """bool: True if the cuckoo filter will expand automatically"""
@@ -184,6 +264,11 @@ class CuckooFilter(object):
         """set the self expand value"""
         self.__auto_expand = bool(val)
 
+    @property
+    def fingerprint_size_bits(self) -> int:
+        """int: The size in bits of the fingerprint"""
+        return self._fingerprint_size
+
     @property
     def fingerprint_size(self) -> int:
         """int: The size in bytes of the fingerprint
@@ -192,7 +277,7 @@ class CuckooFilter(object):
             ValueError: If the size is not between 1 and 4
         Note:
             The size of the fingerprint must be between 1 and 4"""
-        return self.__fingerprint_size // 8
+        return math.ceil(self.fingerprint_size_bits / 8)
 
     @fingerprint_size.setter
     def fingerprint_size(self, val: int):
@@ -202,7 +287,8 @@ class CuckooFilter(object):
             msg = ("{}: fingerprint size must be between 1 and 4").format(self.__class__.__name__)
             raise ValueError(msg)
         # bytes to bits
-        self.__fingerprint_size = tmp * 8  # type: ignore
+        self._fingerprint_size = tmp * 8  # type: ignore
+        self._calc_error_rate()  # if updating fingerprint size then error rate may change
 
     def load_factor(self) -> float:
         """float: How full the Cuckoo Filter is currently"""
@@ -252,7 +338,7 @@ class CuckooFilter(object):
         self._inserted_elements -= 1
         return True
 
-    def export(self, file: typing.Union[Path, str, IOBase, mmap]) -> None:
+    def export(self, file: Union[Path, str, IOBase, mmap]) -> None:
         """Export cuckoo filter to file
 
         Args:
@@ -263,18 +349,12 @@ class CuckooFilter(object):
                 self.export(filepointer)  # type:ignore
         else:
             filepointer = file  # type:ignore
-
             for i in range(len(self.buckets)):
-                bucket = self.buckets[i]
-                # do something for each...
-                if isinstance(bucket, list):
-                    self.buckets[i] = bucket = convert_to_typed(self.__class__.SINGLE_INT_C, bucket)  # type: ignore
-                filepointer.write(bucket.tobytes())  # type: ignore
-                leftover = self.bucket_size - len(bucket)
-                rep = leftover * "I"
-                filepointer.write(pack(rep, *([0] * leftover)))
+                bucket = array(self.SINGLE_INT_C, self.buckets[i])
+                bucket.extend([0] * (self.bucket_size - len(bucket)))
+                bucket.tofile(filepointer)
             # now put out the required information at the end
-            filepointer.write(self.__class__.HEADER_STRUCT.pack(self.bucket_size, self.max_swaps))
+            filepointer.write(self.HEADER_STRUCT.pack(self.bucket_size, self.max_swaps))
 
     def __bytes__(self) -> bytes:
         """Export cuckoo filter to `bytes`"""
@@ -319,9 +399,9 @@ class CuckooFilter(object):
         # if we got here we have an error... we might need to know what is left
         return fingerprint
 
-    def _load(self, file: typing.Union[Path, str, IOBase, mmap]) -> None:
+    def _load(self, file: Union[Path, str, IOBase, mmap, bytes]) -> None:
         """load a cuckoo filter from file"""
-        if not isinstance(file, (IOBase, mmap)):
+        if not isinstance(file, (IOBase, mmap, bytes)):
             file = Path(file)
             with MMap(file) as filepointer:
                 self._load(filepointer)
@@ -332,30 +412,36 @@ class CuckooFilter(object):
             self._parse_buckets(file)  # type: ignore
 
     SINGLE_INT_C = "I"
-    SINGLE_INT_SIZE = calcsize(SINGLE_INT_C)
+    SINGLE_INT_SIZE = Struct(SINGLE_INT_C).size
     HEADER_STRUCT = Struct("II")
 
     def _parse_footer(self, d: ByteString) -> None:
-        list_size = len(d) - self.__class__.HEADER_STRUCT.size
-        self._bucket_size, self.__max_cuckoo_swaps = self.__class__.HEADER_STRUCT.unpack(d[list_size:])  # type:ignore
-        self._cuckoo_capacity = list_size // self.__class__.SINGLE_INT_SIZE // self.bucket_size
+        list_size = len(d) - self.HEADER_STRUCT.size
+        self._bucket_size, self.__max_cuckoo_swaps = self.HEADER_STRUCT.unpack(d[list_size:])  # type:ignore
+        self._cuckoo_capacity = list_size // self.SINGLE_INT_SIZE // self.bucket_size
 
     def _parse_buckets(self, d: ByteString) -> None:
         self._buckets = list()
-        bucket_byte_size = self.bucket_size * self.__class__.SINGLE_INT_SIZE
+        bucket_byte_size = self.bucket_size * self.SINGLE_INT_SIZE
         offs = 0
         for _ in range(self.capacity):
             next_offs = offs + bucket_byte_size
             self.buckets.append(self._parse_bucket(d[offs:next_offs]))  # type: ignore
             offs = next_offs
 
-    def _parse_bucket(self, d: ByteString) -> array.array:
-        bucket = array.ArrayType(self.__class__.SINGLE_INT_C)
-        bucket.frombytes(bytes(d))
-        bucket = convert_to_typed(self.__class__.SINGLE_INT_C, [el for el in bucket if el])
+    def _parse_bucket(self, d: ByteString) -> array:
+        bucket = array(self.SINGLE_INT_C, bytes(d))
+        bucket = array(self.SINGLE_INT_C, [el for el in bucket if el])
         self._inserted_elements += len(bucket)
         return bucket
 
+    def _set_error_rate(self, error_rate: Union[float, None]) -> None:
+        """set error rate correctly"""
+        # if error rate is provided, use it
+        if error_rate is not None:
+            self._error_rate = error_rate
+            self._fingerprint_size = self._calc_fingerprint_size()
+
     def _check_if_present(self, idx_1, idx_2, fingerprint):
         """wrapper for checking if fingerprint is already inserted"""
         if fingerprint in self.buckets[idx_1]:
@@ -409,7 +495,7 @@ class CuckooFilter(object):
         idx_2 = self.__hash_func(str(fingerprint)) % self.capacity
         return idx_1, idx_2
 
-    def _generate_fingerprint_info(self, key: KeyT) -> typing.Tuple[int, int, int]:
+    def _generate_fingerprint_info(self, key: KeyT) -> Tuple[int, int, int]:
         """Generate the fingerprint and indicies using the provided key
 
         Args:
@@ -417,7 +503,7 @@ class CuckooFilter(object):
         """
         # generate the fingerprint along with the two possible indecies
         hash_val = self.__hash_func(key)
-        fingerprint = get_x_bits(hash_val, 64, self.__fingerprint_size, True)
+        fingerprint = get_x_bits(hash_val, 64, self.fingerprint_size_bits, True)
         idx_1, idx_2 = self._indicies_from_fingerprint(fingerprint)
 
         # NOTE: This should never happen...
@@ -435,3 +521,11 @@ class CuckooFilter(object):
         else:
             msg = "The {} is currently full".format(self.__class__.__name__)
             raise CuckooFilterFullError(msg)
+
+    def _calc_error_rate(self):
+        """calculate error rate based on fingerprint size (bits) and bucket size"""
+        return float(1 / (2 ** (self.fingerprint_size_bits - (math.log2(self.bucket_size) + 1))))
+
+    def _calc_fingerprint_size(self) -> int:
+        """calculate fingerprint size (bits) based on error rate and bucket size"""
+        return int(math.ceil(math.log2(1.0 / self.error_rate) + math.log2(self.bucket_size) + 1))
